/**(c)

  Copyright (C) 2006, 2007, 2008, 2009 Michael Stilkerich, Johannes Bauer, Christian Wawersich 

  This file is part of the JOSEK operating system.

	JOSEK is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

	JOSEK is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with JOSEK. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package josek;
import java.util.*;
import java.io.*;

import josek.arch.Architecture;
import josek.conditionparser.ConditionParser;

public class CodeGenerator {
	private HighlevelConfiguration conf;
	private final CommandLineOptions cmdLine;
	private BufferedWriter activeFile = null;
	private int taskNumber = 1;
	private Vector codeHooks;

	CodeGenerator(HighlevelConfiguration conf, CommandLineOptions cmdLine) {
		this.conf = conf;
		this.cmdLine = cmdLine;
		codeHooks = new Vector();
		codeHooks.add((new CodeHook_osh_symbols().init("osh_symbols", conf, cmdLine)));
		codeHooks.add((new CodeHook_osh_prototypes().init("osh_prototypes", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_main_signalhandler().init("osc_main_signalhandler", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_main_preschedule().init("osc_main_preschedule", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_schedule_pretaskhook().init("osc_schedule_pretaskhook", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_schedule_posttaskhook().init("osc_schedule_posttaskhook", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_tasks().init("osc_tasks", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_readyqueues().init("osc_readyqueues", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_counters().init("osc_counters", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_alarms().init("osc_alarms", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_settimer().init("osc_settimer", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_timerhandler().init("osc_timerhandler", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_dumpstack().init("osc_dumpstack", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_ioisr().init("osc_ioisr", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_ArrayOptimizerCode().init("osc_ArrayOptimizerCode", conf, cmdLine)));
		codeHooks.add((new CodeHook_osc_ArrayOptimizerDef().init("osc_ArrayOptimizerDef", conf, cmdLine)));
		codeHooks.add((new CodeHook_osh_ArrayOptimizerDef().init("osh_ArrayOptimizerDef", conf, cmdLine)));
	}

	private void writeLine(String line) throws IOException { writeLine(line, false); }
	protected void writeLine(String line, boolean parse) throws IOException {
		int s,e;

		if(parse) {
			while((s=line.indexOf("//?")) >= 0) {
				if((e=line.indexOf("?//", s)) < 0)
					throw new RuntimeException("Unclosed Variable Substitution");
				String varName = line.substring(s+3, e);
				String varValue = conf.getCodeHookVariable(varName);
				if(varValue==null)
					throw new RuntimeException("Use of unknown Variable: " + varName);
				line = line.substring(0, s) + varValue + line.substring(e+3);
			}
		}
		activeFile.write(line);
		activeFile.write('\n');
	}
	
	protected void writeSingStr() throws IOException {
		activeFile.write(new String(SSingleton.get()));
		activeFile.write('\n');
	}

	private void includeFile(String file, boolean archDependent, boolean parse) throws IOException, FileNotFoundException {
		String independentFile;

		independentFile = cmdLine.getRuleDirectory() + "independent/" + file;
		if (cmdLine.isDefined("codecomments")) writeLine("/**** Code inserted by include " + file + ": ****/");
		if (archDependent) {
			String arch = cmdLine.getArchitecture().toString();
			if (arch == null) {
				throw new RuntimeException("Architecture undefined, but inclusion of architecture dependent file '" + file + "' requested by RULES.");
			}
			file = cmdLine.getRuleDirectory() + arch + "/" + file;
		} else {
			file = independentFile;
		}
		
		BufferedReader includeFile;
		if (file == independentFile) {
			includeFile = openFile(file);
		} else {
			try {
				includeFile = openFile(file);
			} catch (Exception e) {
				includeFile = openFile(independentFile);
			}
		}

		{
		int lineNo=0;
		int nestDepth=-1;
		boolean[] ifTaken = new boolean[16];
		boolean[] eifTaken = new boolean[16];

		for(String line=includeFile.readLine();line!=null;line=includeFile.readLine()) {
			lineNo++;
			if (!parse) { writeLine(line); continue; }

			// IF BLOCK
			if(line.startsWith("//? ")) {
				if(line.length() < 5) { throw new RuntimeException(file+":"+lineNo+": New if-Block lacks condition"); }
				nestDepth++;
				eifTaken[nestDepth]=ifTaken[nestDepth] = ConditionParser.eval(line.substring(4),cmdLine);
				if(nestDepth>0)
					ifTaken[nestDepth] = ifTaken[nestDepth] && ifTaken[nestDepth-1];

				cmdLine.dbgln("IF " + line.substring(4) + " => " + (ifTaken[nestDepth]?"true":"false"),nestDepth);
			
			// END IF BLOCK
			} else if(line.startsWith("//?")) {
				if(nestDepth==-1)
					throw new RuntimeException(file+":"+lineNo+": If-block closed but none opened");
				nestDepth--;
				cmdLine.dbgln("FI => " +((nestDepth==-1 || ifTaken[nestDepth])?"true":"false"),nestDepth);

			// ELSE(-IF) BLOCK
			} else if(line.startsWith("//!")) {
				if(nestDepth==-1) { throw new RuntimeException(file+":"+lineNo+": Else Block without preceeding if Block"); }
				if(eifTaken[nestDepth]) { // earlier block taken
					cmdLine.dbgln("SKIP ELSE (IF)" + line.substring(3));
					ifTaken[nestDepth]=false;
					continue;
				}

				// ELSE-IF BLOCK
				if(line.startsWith("//! ")) {
					 if(line.length() < 5) { throw new RuntimeException(file+":"+lineNo+": New else-if-Block lacks condition"); }
					 cmdLine.dbg("ELSE IF " + line.substring(4) + " => ");
					 eifTaken[nestDepth]=ifTaken[nestDepth] = ConditionParser.eval(line.substring(4),cmdLine);

				// ELSE BLOCK
				} else {
					ifTaken[nestDepth] = true;
					cmdLine.dbg("ELSE => ");
				}
				if(nestDepth>0)
					ifTaken[nestDepth] = ifTaken[nestDepth] && ifTaken[nestDepth-1];
				cmdLine.dbgln((ifTaken[nestDepth]?"true":"false"),nestDepth);

			// INCLUDE STATEMENT
			} else if(line.startsWith("//+")) {
				if(nestDepth==-1 || ifTaken[nestDepth])
					activateHook(line.substring(4), file);

			// REGULAR LINE
			} else if (nestDepth==-1 || ifTaken[nestDepth]) {
				try {
					writeLine(line, true);
				} catch (RuntimeException e) {
					throw new RuntimeException(file+":"+lineNo+": "+e.getMessage());
				}
			}
		}
		if(nestDepth >= 0)
			throw new RuntimeException(file+":"+lineNo+": EOF reached with unclosed ifblocks");
		}
		includeFile.close();
		if (cmdLine.isDefined("codecomments")) writeLine("/**** Code insertion done. ****/");
	}

	private void activateHook(String hook, String file) {
		for (int i = 0; i < codeHooks.size(); i++) {
			CodeHook c = (CodeHook)codeHooks.get(i);
			if (((CodeHook)codeHooks.get(i)).fire(hook)) return;
		}
		throw new RuntimeException("Unknown hook " + hook + " specified, included from " + file + ".");
	}

	private static BufferedReader openFile(String filename) throws FileNotFoundException {
		BufferedReader br;
		InputStream is = ClassLoader.getSystemResourceAsStream(filename);

		if(is == null) {
			br = new BufferedReader(new FileReader(filename));
		} else {
			br = new BufferedReader(new InputStreamReader (is));
		}
		
		return br;
	}

	public void generate() throws IOException, FileNotFoundException {
		String file = cmdLine.getRuleDirectory() + "RULES";
		BufferedReader rulesFile = openFile(file);

		int lineNo = 0;
		int nestDepth=-1;
		boolean parseForInternalHooks;
		boolean[] ifTaken = new boolean[16];
		boolean[] eifTaken = new boolean[16];

		for(String line=rulesFile.readLine();line!=null;line=rulesFile.readLine()) {
			lineNo++;

			StringTokenizer tok = new StringTokenizer(line, " \t");
			String tokens[] = new String[2];
			for (int i = 0; i < tokens.length; i++)
				if (tok.hasMoreTokens()) tokens[i] = tok.nextToken();
			if(tokens[0]==null) continue; // Whitespace line

			char lineType = tokens[0].charAt(0);
			if(nestDepth>=0 && !ifTaken[nestDepth] && lineType!='?' && lineType!='@')
				continue;
			
			switch (lineType) {
				case '#':
					break;

				case ':': 
					if (activeFile != null) activeFile.close();
					activeFile = new BufferedWriter(new FileWriter(cmdLine.getOutputDirectory() + tokens[1]));
					for (int i = 0; i < codeHooks.size(); i++) ((CodeHook)codeHooks.get(i)).changedFile(activeFile);
					break;

				case '~': 
					parseForInternalHooks = (tokens[0].length() >= 2) && (tokens[0].charAt(1) == '+');
					includeFile(tokens[1], false, parseForInternalHooks);
					break;

				case '!': 
					parseForInternalHooks = (tokens[0].length() >= 2) && (tokens[0].charAt(1) == '+');
					includeFile(tokens[1], true, parseForInternalHooks);
					break;

				case '%': 
					activateHook(tokens[1], "RULES");
					break;

				// ELSE(-IF) BLOCK
				case '@':
					if(nestDepth==-1) { throw new RuntimeException(file+":"+lineNo+": Else Block without preceeding if Block"); }
					if(eifTaken[nestDepth]) { // earlier block taken
						cmdLine.dbgln("SKIP ELSE (IF)" + line.substring(1));
						ifTaken[nestDepth]=false;
						break;
					}

					// ELSE-IF BLOCK
					if (tokens[1] != null) {
						cmdLine.dbg("ELSE IF " + line.substring(2) + " => ");
						eifTaken[nestDepth]=ifTaken[nestDepth] = ConditionParser.eval(line.substring(2),cmdLine);

					// ELSE BLOCK
					} else {
						ifTaken[nestDepth] = true;
						cmdLine.dbg("ELSE => ");
					}
					if(nestDepth>0)
						ifTaken[nestDepth] = ifTaken[nestDepth] && ifTaken[nestDepth-1];
					cmdLine.dbgln((ifTaken[nestDepth]?"true":"false"),nestDepth);
					break;

					// (END) IF BLOCK
				case '?': 
					if (tokens[1] == null) { // End IF
						if(nestDepth==-1)
							throw new RuntimeException(file+":"+lineNo+": If-block closed but none opened");
						nestDepth--;
						cmdLine.dbgln("FI => " +((nestDepth==-1 || ifTaken[nestDepth])?"true":"false"),nestDepth);
					} else { // Begin IF
						nestDepth++;
						eifTaken[nestDepth]=ifTaken[nestDepth] = ConditionParser.eval(line.substring(2),cmdLine);
						if(nestDepth>0)
							ifTaken[nestDepth] = ifTaken[nestDepth] && ifTaken[nestDepth-1];
						cmdLine.dbgln("IF " + line.substring(2) + " => " + (ifTaken[nestDepth]?"true":"false"),nestDepth);
					}
					break;

				default: 
					throw new RuntimeException("Unknown token '" + lineType + "' in line " + lineNo + " of rules file.");

			}
		}
		try {
			if (activeFile != null) activeFile.close();
		} catch (IOException e) {
			throw new RuntimeException("IO-Error closing last file specified in rules: " + e);
		}
	}
}

