/**(c)

  Copyright (C) 2006, 2007, 2008, 2009 Michael Stilkerich, Johannes Bauer, Christian Wawersich 

  This file is part of the JOSEK operating system.

	JOSEK is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

	JOSEK is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with JOSEK. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package josek;

import java.util.Vector;
import java.util.HashSet;

import josek.arch.Architecture;

final public class CommandLineOptions {

	private	boolean acceptUnknownAttributes;
	private	Vector includePaths;
	private	String commandFile;
	private	boolean generateResourceStatistics;
	private	boolean testVerify;
	private	int verbose;
	private	String ruleDirectory;
	private	String outputDirectory;
	private	Architecture architecture;
	private	HashSet defines;
	private	HashSet validSymbols;

	private void verb(String msg, int indent, int level) {
		if(verbose>=level) {
			while(indent-->0)
				System.err.print("\t");
			System.err.print(msg);
		}
	}

	private void verbln(String msg, int indent, int level) {
		if(verbose>=level) {
			verb(msg,indent);
			System.err.println();
		}
	}

	public void dbgln(String msg, int indent) {
		verbln(msg,indent,2);
	}

	public void dbg(String msg, int indent) {
		verb(msg,indent,2);
	}

	public void verbln(String msg, int indent) {
		verbln(msg,indent,1);
	}

	public void verb(String msg, int indent) {
		verb(msg,indent,1);
	}

	public void verb(String msg) { verb(msg,0, 1); }
	public void verbln(String msg) { verbln(msg,0, 1); }
	public void dbg(String msg) { verb(msg,0,2); }
	public void dbgln(String msg) { verbln(msg,0,2); }

	public String getCommandFile() {
		return commandFile;
	}

	public String getRuleDirectory() {
		return ruleDirectory;
	}

	public String getOutputDirectory() {
		return outputDirectory;
	}

	public boolean getGenerateResourceStatistics() {
		return generateResourceStatistics;
	}

	public boolean isDefined(String symbol) {
		if (!validSymbols.contains(symbol)) throw new RuntimeException("Query for undocumented symbol '" + symbol + "'");
		return defines.contains(symbol);
	}

	public void defineSymbol(String symbol) {
		if (!validSymbols.contains(symbol)) throw new RuntimeException("Define of undocumented symbol '" + symbol + "'");
		defines.add(symbol);
	}

	public Architecture getArchitecture() {
		return architecture;
	}

	private int parseSingleArgument(String[] args, int index) {
		if (args[index].equals("-a")) {
			acceptUnknownAttributes = true;
			return 0;
		} else if (args[index].equals("-i")) {
			includePaths.addElement(args[index + 1]);
			return 1;
		} else if (args[index].equals("-f")) {
			commandFile = args[index + 1];
			return 1;
		} else if (args[index].equals("-r")) {
			generateResourceStatistics = true;
			return 0;
		} else if (args[index].equals("-V")) {
			System.err.println("Version: 0.01");
			System.exit(0);
			return 0;
		} else if (args[index].equals("-t")) {
			testVerify = true;
			return 0;
		} else if (args[index].equals("-v")) {
			verbose++;
			return 0;
		} else if (args[index].equals("-c")) {
			ruleDirectory = args[index + 1];
			return 1;
		} else if (args[index].equals("-o")) {
			outputDirectory = args[index + 1];
			return 1;
		} else if (args[index].equals("-D")) {
			defineSymbol(args[index + 1]);
			return 1;
		} else if (args[index].equals("-h")) {
			try {
				architecture = (Architecture) Class.forName("josek.arch." + args[index + 1]).newInstance();
				architecture.setCommandLine(this);
				architecture.setArchOptions();
			} catch (Exception e) {
				System.err.println("ERROR: Architecture \"" +args[index + 1] + "\" not supported!");
				throw new RuntimeException(e);
			}
			return 1;
		}
		displayHelp();
		throw new RuntimeException("Unparsable command line option (#" + (index + 1) + "): " + args[index]);
	}

	public CommandLineOptions(String[] args) {
		acceptUnknownAttributes = false;
		includePaths = new Vector();
		commandFile = null;
		generateResourceStatistics = false;
		testVerify = false;
		verbose = 0;
		outputDirectory = "generated/";
		ruleDirectory = "josek/code/arch/";
		defines = new HashSet();
		validSymbols = new HashSet();
		validSymbols.add("assertions");
		validSymbols.add("codecomments");
		validSymbols.add("debug");
		validSymbols.add("have_alarms");
		validSymbols.add("have_ecc");
		validSymbols.add("have_extstatus");
		validSymbols.add("have_ioisr");
		validSymbols.add("have_linux");
		validSymbols.add("have_multi_act");
		validSymbols.add("have_posttask_hook");
		validSymbols.add("have_preemption");
		validSymbols.add("have_pretask_hook");
		validSymbols.add("have_readyqueues");
		validSymbols.add("have_resources");
		validSymbols.add("have_stdlib");
		validSymbols.add("have_task_flags");
		validSymbols.add("have_tasks");
		validSymbols.add("mixed_scheduling");
		validSymbols.add("saveram");
		validSymbols.add("signals");
		validSymbols.add("termidle");
		validSymbols.add("timer");

		if (args.length == 0) {
			displayHelp();
			throw new RuntimeException("No command line parameters given.");
		}

		for (int i = 0; i < args.length; i++) {
			int j = parseSingleArgument(args, i);
			i += j;
		}

		if (getArchitecture() == null) {
			try {
				architecture = (Architecture) Class.forName("josek.arch.avr").newInstance();
				architecture.setCommandLine(this);
				architecture.setArchOptions();
			} catch (Exception e) {
				System.err.println("Error: No architecture specified!");
				throw new RuntimeException(e);
			}
			System.err.println("warning: Architecture is not specified! Use avr as default");
		}
	}

	void displayHelp() {
		System.err.println("usage: josek [-a] [-h arch] [-i path] [-f file] [-r] [-V] [-v] [-t]");
		System.err.println("    -a         Accept unknown attributes");
		System.err.println("    -i Path    Include 'Path' in search for includes");
		System.err.println("    -f File    Use 'File' as command file");
		System.err.println("    -r         Generate resource statistics");
		System.err.println("    -V         Display version number");
		System.err.println("    -v         Generate verbose output");
		System.err.println("    -t         Enable test/verification mode");
		System.err.println("    -h Arch    Define architecture (x86_32/x86_64/avr)");
		System.err.println("    -c Path    Specify rule directory");
		System.err.println("    -o Path    Specify output directory");
		System.err.println("    -D Symbol  Define symbol 'Symbol'");
		System.err.println();
		System.err.println("Known symbols (and the results on the generated code):");
		System.err.println("    signals        Signal handler functions");
		System.err.println("    codecomments   Comments from where the included code originates");
		System.err.println("    debug          Debugging functions (dump_xxx)");
		System.err.println("    assertions     Include assertions in code");
		System.err.println("    timer          Include code for timers");
		System.err.println("    saveram        Preserve RAM by generating static lookup tables");
		System.err.println("    termidle       ShutdownOS() when system is idle");
	}
}
