/**(c)

  Copyright (C) 2006, 2007, 2008, 2009 Michael Stilkerich, Johannes Bauer, Christian Wawersich 

  This file is part of the JOSEK operating system.

	JOSEK is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

	JOSEK is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with JOSEK. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package josek;

import josek.idmap.IDMap;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Enumeration;

final public class HighlevelConfiguration {
	private Hashtable<String, HighlevelConfiguration_Resource> resources;
	private boolean haveSchedulerResource;
	private Vector<HighlevelConfiguration_Event> events;

	public HighlevelConfiguration_Hooks hooks;

	private IDMap<HighlevelConfiguration_Alarm, HighlevelConfiguration_Counter> mapAlarmCounter;
	private IDMap<HighlevelConfiguration_Task, Priority> mapTaskPriority;

	private Vector<HighlevelConfiguration_ISR> ISRs;
	public ISRLevels ISRLevels;
	private int IOISRCount;
	private int timerInterval;
	private Vector<ArrayOptimizer> arrayOptimizers;
	private final CommandLineOptions cmdLine;
	private final Hashtable<String,String> codeHookVariables; 
	
	public void addArrayOptimizer(ArrayOptimizer a) {
		arrayOptimizers.add(a);
	}

	public ArrayOptimizer getArrayOptimizer(int i) {
		return (ArrayOptimizer)arrayOptimizers.get(i);
	}

	public int getArrayOptimizerCount() {
		return arrayOptimizers.size();
	}

	/**
	 * Returns the number of OSEK Priorities.
	 * There is one more internal priority that is reserved for the idle
	 * task.
	 */
	int getNumPriorities() {
		return mapTaskPriority.getBaseCount();
	}

	public int getTaskCount() {
		return mapTaskPriority.getChildCount();
	}

	public IDMap<HighlevelConfiguration_Task, Priority> taskPrioMapper() {
		return mapTaskPriority;
	}

	public int getMaxTaskID() {
		return mapTaskPriority.getMaxChildID();
	}

	public int getResourceCount() {
		return resources.size();
	}

	public int getEventCount() {
		return events.size();
	}

	public HighlevelConfiguration_Event getEvent(int i) {
		return (HighlevelConfiguration_Event)events.get(i);
	}

	public int getCounterCount() {
		return mapAlarmCounter.getBaseCount();
	}

	public int getMaxCounterID() {
		return mapAlarmCounter.getMaxBaseID();
	}

	public int getMaxAlarmID() {
		return mapAlarmCounter.getMaxChildID();
	}

	public Iterator<HighlevelConfiguration_Counter> getCounterIterator() {
		return mapAlarmCounter.baseIterator();
	}

	public Iterator<Priority> getPriorityIterator() {
		return mapTaskPriority.baseIterator();
	}

	public Enumeration<HighlevelConfiguration_Task> getTasks() {
		return mapTaskPriority.childElements();
	}

	Enumeration<HighlevelConfiguration_Alarm> getAlarms() {
		return mapAlarmCounter.childElements();
	}

	Enumeration<HighlevelConfiguration_Resource> getResources() {
		return resources.elements();
	}

	public String getAlarmCounterRevMapMacro() {
		return mapAlarmCounter.getReverseMapMacro();
	}

	public String getTaskPriorityRevMapMacro() {
		return mapTaskPriority.getReverseMapMacro();
	}
	public int getAlarmCount() {
		return mapAlarmCounter.getChildCount();
	}

	public int getISRCount() {
		return ISRs.size();
	}

	public HighlevelConfiguration_ISR getISR(int i) {
		return (HighlevelConfiguration_ISR)ISRs.get(i);
	}

	public boolean haveIOISRs() {
		return IOISRCount > 0;
	}
	
	public int getIOISRCount() {
		return IOISRCount;
	}

	public int getTimerInterval() {
		return timerInterval;
	}

	private void parseTask(PrimitiveStringAttribute a) {
		String name = a.value;
		boolean autoActivation = false;
		int priority = 0;
		int mact=1;
		int stacksize = 65536;
		int sched_model = 0;
		boolean preempt = cmdLine.getArchitecture().supportsPreemption();
		Vector<String> usedResources = new Vector<String>();
		Vector<String> usedEvents = new Vector<String>();

		for (int i = 0; i < a.children.size(); i++) {
			Attribute node = (Attribute)a.children.get(i);
			if (node instanceof PrimitiveBoolAttribute) {
				PrimitiveBoolAttribute bnode = (PrimitiveBoolAttribute)node;
				if (bnode.lValue.equals("AUTOSTART")) autoActivation = bnode.value;
			} else if (node instanceof PrimitiveIntAttribute) {
				PrimitiveIntAttribute inode = (PrimitiveIntAttribute)node;
				if (inode.lValue.equals("PRIORITY")) priority = inode.value;
				if (inode.lValue.equals("STACKSIZE")) stacksize = inode.value;
				if (inode.lValue.equals("ACTIVATION")) {
					mact = inode.value;
					if(mact>1) cmdLine.defineSymbol("have_multi_act");
				}
			} else if (node instanceof PrimitiveStringAttribute) {
				PrimitiveStringAttribute snode = (PrimitiveStringAttribute)node;
				if (snode.lValue.equals("RESOURCE")) usedResources.add(snode.value);
				else if (snode.lValue.equals("EVENT")) usedEvents.add(snode.value);
				else if (snode.lValue.compareTo("SCHEDULE")==0) { // NON or FULL
					if(snode.value.compareTo("FULL")==0) {
						preempt=true;
						cmdLine.defineSymbol("have_preemption");
					} else if(snode.value.compareTo("NON")==0) {
						preempt=false;
					} else {
						throw new RuntimeException("Unknown scheduling mode " + snode.value);
					}
				}
			}
		}

		if(preempt && !cmdLine.getArchitecture().supportsPreemption())
				throw new RuntimeException("Architecture "+cmdLine.getArchitecture().toString()+" does not support preemption, but requested from OIL file");

		Priority pobj;
		String pname = "Prio" + priority;
		if(!mapTaskPriority.existsBase(pname)) {
			pobj = new Priority(priority, mact, pname);
			mapTaskPriority.addBase(pobj);
		} else {
			pobj =  mapTaskPriority.getBase(pname);
			pobj.incActivations(mact);
		}

		HighlevelConfiguration_Task t = new HighlevelConfiguration_Task(name, autoActivation, priority, stacksize, mact, preempt, usedResources, usedEvents);
		mapTaskPriority.addChild(t, pname);

		cmdLine.defineSymbol("have_tasks");
	}

	private void parseOS(PrimitiveStringAttribute a) {
		for (int i = 0; i < a.children.size(); i++) {
			Attribute node = (Attribute)a.children.get(i);
			if (node instanceof PrimitiveBoolAttribute) {
				PrimitiveBoolAttribute bnode = (PrimitiveBoolAttribute)node;
				if (bnode.lValue.equals("PRETASKHOOK")) {
					hooks.preTask = bnode.value;
					if(bnode.value) cmdLine.defineSymbol("have_pretask_hook");
				}	else if (bnode.lValue.equals("POSTTASKHOOK")) {
					hooks.postTask = bnode.value;
					if(bnode.value) cmdLine.defineSymbol("have_posttask_hook");
				}	else if (bnode.lValue.equals("STARTUPHOOK")) {
					hooks.startup = bnode.value;
				}	else if (bnode.lValue.equals("SHUTDOWNHOOK")) {
					hooks.shutdown = bnode.value;
				}		else if (bnode.lValue.equals("ERRORHOOK")) {
					hooks.error = bnode.value;
				}			else if (bnode.lValue.equals("PREISRHOOK")) {
					hooks.preISR = bnode.value;
				}				else if (bnode.lValue.equals("POSTISRHOOK")) {
					hooks.postISR = bnode.value;
				}					else if (bnode.lValue.equals("USERESSCHEDULER")) {
					haveSchedulerResource = bnode.value;
					if( haveSchedulerResource ) {
						cmdLine.defineSymbol("have_resources");
					}
				}
			} else if (node instanceof PrimitiveStringAttribute) {
				PrimitiveStringAttribute snode = (PrimitiveStringAttribute)node;
				if (snode.lValue.equals("STATUS") && snode.value.equals("EXTENDED")) cmdLine.defineSymbol("have_extstatus");
				if (snode.lValue.equals("MCU") || snode.lValue.equals("CPU")|| snode.lValue.equals("MICROCONTROLLER")) setCodeHookVariable("mcu", snode.value.toLowerCase());
			} else if (node instanceof PrimitiveIntAttribute) {
				PrimitiveIntAttribute inode = (PrimitiveIntAttribute)node;
				if (inode.lValue.equals("FREQ")) setCodeHookVariable("mcufreq", Integer.toString(inode.value));
			}
		}
	}

	private void parseResource(PrimitiveStringAttribute a) {
		HighlevelConfiguration_Resource r = new HighlevelConfiguration_Resource(a.value);
		cmdLine.defineSymbol("have_resources");
		resources.put(a.value, r);
	}
	
	private void parseEvent(PrimitiveStringAttribute a) {
		int mask = 0;
		if(!cmdLine.getArchitecture().supportsECC()) {
			throw new RuntimeException("Events specified in OIL but ECC is not supported on architecture " +
					cmdLine.getArchitecture().toString());
		}
		cmdLine.defineSymbol("have_ecc");
		for (int i = 0; i < a.children.size(); i++) {
			Attribute node = (Attribute)a.children.get(i);
			if (node instanceof PrimitiveIntAttribute) {
				PrimitiveIntAttribute inode = (PrimitiveIntAttribute)node;
				if (inode.lValue.equals("MASK")) mask = inode.value;
			}
		}
		HighlevelConfiguration_Event e = new HighlevelConfiguration_Event(a.value, mask);
		events.add(e);
	}
	
	private void parseCounter(PrimitiveStringAttribute a) {
		int cntrMaxValue = 100;
		int minCycle = 1;
		int ticksPerBase = 1;
		int timeInNanoSeconds = 1000000000;
		String base = null;
		for (int i = 0; i < a.children.size(); i++) {
			Attribute node = (Attribute)a.children.get(i);
			if (node instanceof PrimitiveIntAttribute) {
				PrimitiveIntAttribute inode = (PrimitiveIntAttribute)node;
				if (inode.lValue.equals("MAXALLOWEDVALUE")) cntrMaxValue = inode.value;
					else if (inode.lValue.equals("MINCYCLE")) minCycle = inode.value;
					else if (inode.lValue.equals("TICKSPERBASE")) ticksPerBase = inode.value;
					else if (inode.lValue.equals("TIME_IN_NS")) timeInNanoSeconds = inode.value;
			} else if (node instanceof PrimitiveStringAttribute) {
				PrimitiveStringAttribute snode = (PrimitiveStringAttribute) node;
				if (snode.lValue.equals("TIMER_BASE")) base = snode.value;
			}
		}
		HighlevelConfiguration_Counter c = new HighlevelConfiguration_Counter(a.value, cntrMaxValue, minCycle, ticksPerBase, timeInNanoSeconds, base);
		mapAlarmCounter.addBase(c);
	}
	
	private void parseAlarm(PrimitiveStringAttribute a) {
		String counter = null;
		boolean autoStart = false;
		AlarmEvent event = null;
		int alarmTime = 0;
		int cycleTime = 0;

		for (int i = 0; i < a.children.size(); i++) {
			Attribute node = (Attribute)a.children.get(i);
			if (node instanceof PrimitiveBoolAttribute) {
				PrimitiveBoolAttribute bnode = (PrimitiveBoolAttribute)node;
				if (bnode.lValue.equals("AUTOSTART")) {
					autoStart = bnode.value;
					for (int j = 0; j < bnode.children.size(); j++) {
						Attribute innerNode = (Attribute)bnode.children.get(j);
						if (innerNode instanceof PrimitiveIntAttribute) {
							PrimitiveIntAttribute innerIntNode = (PrimitiveIntAttribute)innerNode;
							if (innerIntNode.lValue.equals("ALARMTIME")) alarmTime = innerIntNode.value;
								else if (innerIntNode.lValue.equals("CYCLETIME")) cycleTime = innerIntNode.value;
						}
					}
				}
			} else if (node instanceof PrimitiveStringAttribute) {
				PrimitiveStringAttribute snode = (PrimitiveStringAttribute)node;
				if (snode.lValue.equals("ACTION")) {
					if (snode.value.equals("SETEVENT")) {
						String eventName = null;
						String taskName = null;
						for (int j = 0; j < snode.children.size(); j++) {
							Attribute innernode = (Attribute)snode.children.get(j);
							if (innernode instanceof PrimitiveStringAttribute) {
								PrimitiveStringAttribute innersnode = (PrimitiveStringAttribute)innernode;
								if (innersnode.lValue.equals("TASK")) taskName = innersnode.value;
									else if (innersnode.lValue.equals("EVENT")) eventName = innersnode.value;
							}
						}
						event = new AlarmEvent_SetEvent(eventName, taskName);
					} else if (snode.value.equals("ACTIVATETASK")) {
						String taskName = null;
						for (int j = 0; j < snode.children.size(); j++) {
							Attribute innernode = (Attribute)snode.children.get(j);
							if (innernode instanceof PrimitiveStringAttribute) {
								PrimitiveStringAttribute innersnode = (PrimitiveStringAttribute)innernode;
								if (innersnode.lValue.equals("TASK")) taskName = innersnode.value;
							}
						}
						event = new AlarmEvent_ActivateTask(taskName);
					} else if (snode.value.equals("ALARMCALLBACK")) {
						String funcName = null;
						for (int j = 0; j < snode.children.size(); j++) {
							Attribute innernode = (Attribute)snode.children.get(j);
							if (innernode instanceof PrimitiveStringAttribute) {
								PrimitiveStringAttribute innersnode = (PrimitiveStringAttribute)innernode;
								if (innersnode.lValue.equals("ALARMCALLBACKNAME")) funcName = innersnode.value;
							}
						}
						event = new AlarmEvent_Callback(funcName);
					} else {
						throw new RuntimeException("Unknown action type for alarm '" + a.value + "': " + snode.value);
					}
				} else if (snode.lValue.equals("COUNTER")) {
					counter = snode.value;
				}
			}
		}
	
		if(autoStart && alarmTime==0) {
			throw new RuntimeException("AlarmTime needs to be specified for autostarted Alarm '" + a.value + "'");
		}

		HighlevelConfiguration_Alarm alrm = new HighlevelConfiguration_Alarm(a.value, counter, alarmTime, cycleTime, autoStart, event);
		mapAlarmCounter.addChild(alrm, counter);
		cmdLine.defineSymbol("have_alarms");
	}
	
	private void parseISR(PrimitiveStringAttribute a) {
		int category = 2;
		int priority = 99;
		int stackSize = 0;
		int signal = -1;
		int irqLevel = -1;
		for (int i = 0; i < a.children.size(); i++) {
			Attribute node = (Attribute)a.children.get(i);
			if (node instanceof PrimitiveIntAttribute) {
				PrimitiveIntAttribute inode = (PrimitiveIntAttribute)node;
				if (inode.lValue.equals("CATEGORY")) category = inode.value;
					else if (inode.lValue.equals("PRIORITY")) priority = inode.value;
					else if (inode.lValue.equals("STACKSIZE")) stackSize = inode.value;
					else if (inode.lValue.equals("IRQLEVEL")) irqLevel = inode.value;
			} else if (node instanceof PrimitiveStringAttribute) {
				PrimitiveStringAttribute snode = (PrimitiveStringAttribute)node;
				if (snode.lValue.equals("SIGNAL")) {
					if (snode.value.equals("SIGUSR1")) signal = 10;
						else if (snode.value.equals("SIGUSR2")) signal = 12;
						else if (snode.value.equals("IO")) signal = 0;
				}
			}
		}
		if (signal == -1 && cmdLine.getArchitecture().hasLinux()) {
			throw new RuntimeException("Unknown or undefined signal for ISR '" + a.value);
		}
		HighlevelConfiguration_ISR i = new HighlevelConfiguration_ISR(a.value, category, priority, stackSize, signal, irqLevel);
		ISRs.add(i);
	}
	
	private boolean inspect(Attribute node, int level) {
		if (level == 2) {
			if (node instanceof PrimitiveStringAttribute) {
				PrimitiveStringAttribute a = (PrimitiveStringAttribute)node;
				if (a.lValue.equals("TASK")) {
					parseTask(a);
					return false;
				} else if (a.lValue.equals("OS")) {
					parseOS(a);
					return false;
				} else if (a.lValue.equals("RESOURCE")) {
					parseResource(a);
					return false;
				} else if (a.lValue.equals("EVENT")) {
					parseEvent(a);
					return false;
				} else if (a.lValue.equals("COUNTER")) {
					parseCounter(a);
					return false;
				} else if (a.lValue.equals("ALARM")) {
					parseAlarm(a);
					return false;
				} else if (a.lValue.equals("ISR")) {
					parseISR(a);
					return false;
				}
			}
		}
		return true;
	}

	private void parse(Attribute node, int level) {
		if (inspect(node, level)) {
			for (int i = 0; i < node.children.size(); i++) {
				parse((Attribute)node.children.get(i), level + 1);
			}
		}
	}

	public String getCodeHookVariable(String key) {
		return (String) codeHookVariables.get(key);
	}
	public void setCodeHookVariable(String key, String val) {
		codeHookVariables.put(key,val);
	}

	HighlevelConfiguration(Configuration conf, CommandLineOptions cmdLine) {
		this.hooks = new HighlevelConfiguration_Hooks();
		this.resources = new Hashtable<String, HighlevelConfiguration_Resource>();
		this.events = new Vector<HighlevelConfiguration_Event>();
		this.ISRs = new Vector<HighlevelConfiguration_ISR>();
		this.arrayOptimizers = new Vector<ArrayOptimizer>();
		this.codeHookVariables = new Hashtable<String, String>();
		this.cmdLine = cmdLine;
		this.mapAlarmCounter = new IDMap<HighlevelConfiguration_Alarm, HighlevelConfiguration_Counter>(0, "Counter", "Alarm", cmdLine,this);
		this.mapTaskPriority = new IDMap<HighlevelConfiguration_Task, Priority>(0, "Priority", "Task", cmdLine, this); /* Prio 0 is reserved for the Idle Task */
		haveSchedulerResource = false;
		cmdLine.getArchitecture().setHighlevelConfig(this);

		parse(conf.getRoot(), 0);

		/* Mapping Priorities <=> Tasks, create internal priority numbers */
		mapTaskPriority.map();

		// FIXME choose a suitable readyplist type
		if(getNumPriorities() > cmdLine.getArchitecture().machineWordSize()) { 
			throw new RuntimeException("Only" + cmdLine.getArchitecture().machineWordSize() +
					"Priority levels supported currently");
		}

		/* Add scheduler resource */
		if(haveSchedulerResource) {
			HighlevelConfiguration_Resource r = new HighlevelConfiguration_Resource("RES_SCHEDULER");
			r.setUsedByPriority(0); // 0 is the highest internal priority
			System.err.println("RESSCHEDULER CP = " + mapTaskPriority.getMaxBaseID());
			resources.put("RES_SCHEDULER", r);
		}

		/* Now calculate ceiling priorities */
		for(Enumeration<HighlevelConfiguration_Task> e=mapTaskPriority.childElements();
				e.hasMoreElements(); ) {
			HighlevelConfiguration_Task tobj = e.nextElement();

			for (int j = 0; j < tobj.getUsedResourceCount(); j++) {
				String resName = tobj.getUsedResource(j);
				HighlevelConfiguration_Resource robj = resources.get(resName);
				if(robj==null) {
					throw new RuntimeException("Unknown referenced resource " + resName);
				}
				robj.setUsedByPriority(tobj.getBase().getInternalID());
			}
		}
		for(Enumeration<HighlevelConfiguration_Resource> re=resources.elements();
				re.hasMoreElements(); ) {
			HighlevelConfiguration_Resource robj = re.nextElement();
			int internalCP = robj.getCeilingPriority();
			mapTaskPriority.getBaseByIID(internalCP).setIsCeilingPrio();
		}

		// do we have mixed scheduling?
		if(HighlevelConfiguration_Task.getSchedMode() == 3) {
			cmdLine.defineSymbol("mixed_scheduling");
			cmdLine.defineSymbol("have_task_flags");
		}

		// We need readyqueues if we have more than one task per prio
		if(mapTaskPriority.getMaxCPB() > 1) {
			cmdLine.defineSymbol("have_readyqueues");
		}

		/* Reverse mapping of Counters <=> Alarms */
		mapAlarmCounter.map();

		if (getCounterCount() > 0) {
			cmdLine.defineSymbol("timer");
			if(cmdLine.getArchitecture().hasLinux())
				cmdLine.defineSymbol("signals");
			codeHookVariables.put("TickType", HighlevelConfiguration_Counter.getTickType()); 
		}

		/* Do we have any IO interrupts? */
		for (int i = 0; i < getISRCount(); i++) {
			if (getISR(i).getSignal() == 0) {
				IOISRCount++;
			}
		}
		if (IOISRCount > 0) {
			cmdLine.defineSymbol("have_ioisr");
			cmdLine.defineSymbol("signals");
		}

		/* Configure interrupt levels */
		if(cmdLine.getArchitecture().hasISRLevels())
			ISRLevels = new ISRLevels(this);
		
		/* Optimizations on arrays must come last */
		AddArrayOptimizers.work(this, cmdLine);
	}

	void dumpConfiguration() {
		System.out.println("There are " + getTaskCount() + " tasks in the system:");
		for (Iterator<Priority> it=mapTaskPriority.baseIterator(); it.hasNext(); ) {
			Priority pobj = it.next();
			for (int i = 0; i < pobj.numChildren(); i++) {
				System.out.print("Priority " + pobj.getInternalID() + " ");
				System.out.println(pobj.getChild(i));
			}
		}
		System.out.println();
		
		System.out.println("There are " + getResourceCount() + " resources in the system:");
		for(Enumeration <HighlevelConfiguration_Resource> e = resources.elements();
				e.hasMoreElements(); ) {
			System.out.println("    " + e.nextElement());
		}
		System.out.println();

		System.out.println("There are " + getEventCount() + " events in the system:");
		for (int i = 0; i < getEventCount(); i++) {
			System.out.println("    " + getEvent(i));
		}
		System.out.println();
		
		System.out.println("There are " + getCounterCount() + " counters in the system (time = " + getTimerInterval() + "):");
		for (Iterator<HighlevelConfiguration_Counter> it = mapAlarmCounter.baseIterator(); it.hasNext(); ) {
			System.out.println("    " + it.next());
		}
		System.out.println();
		
		System.out.println("There are " + getAlarmCount() + " alarms in the system:");
		for (Iterator<HighlevelConfiguration_Counter> it = mapAlarmCounter.baseIterator(); it.hasNext(); ) {
			HighlevelConfiguration_Counter ctr = it.next();

			for (int i = 0; i < ctr.numChildren(); i++) {
				System.out.println("    " + ctr.getChild(i));
			}
		}
		System.out.println();
		
		System.out.println(getISRCount() + " ISRs in the system:");
		for (int i = 0; i < getISRCount(); i++) {
			System.out.println("    " + getISR(i));
		}
		System.out.println();
	}
}

