/**(c)

  Copyright (C) 2006, 2007, 2008, 2009 Michael Stilkerich, Johannes Bauer, Christian Wawersich 

  This file is part of the JOSEK operating system.

	JOSEK is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

	JOSEK is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with JOSEK. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package josek.arch.atmega;

import josek.HighlevelConfiguration;
import josek.HighlevelConfiguration_Counter;
import josek.CommandLineOptions;
import josek.arch.avr;

public abstract class ATMega {
	protected CommandLineOptions cmdLine;
	protected HighlevelConfiguration conf;
	protected avr arch;
	int[] pscalers= { 1, 8, 64, 256, 1024 };
	int[] epscalers= { 1, 8, 32, 64, 128, 256, 1024 };
	int[][] timer_prescalers;
	int[] timer_max;
	boolean[] counters_used;

	int timsk,etimsk;

	// Counter Regs
	int[] tccr;
	int[] tccrb;
	int[] ocr;
	ATMega() {
		tccr = new int[4];
		tccrb = new int[4];
		ocr = new int[4];
	}
	long maxNanoSeconds(String timer_base) {
		int ti = tind(timer_base);
		return maxNanoSeconds(ti);
	}
	long maxNanoSeconds(int ti, int pindex) {
		int psclr = timer_prescalers[ti][pindex];

		// freq is in Hz
		long max = timer_max[ti] * 1000000000L; // ns
		long slowestFreq = arch.freq() / psclr;
		max /= slowestFreq;
		return max;
	}
	long maxNanoSeconds(int ti) {
		return maxNanoSeconds(ti, timer_prescalers[ti].length-1);
	}
	long minNanoSeconds(String timer_base) {
		long min = 1000000000L/(long) arch.freq();
		// in real-life, an interrupt per cycle would obviously not work :-)
		return min;
	}

	public void addhandler(StringBuffer s, HighlevelConfiguration_Counter ctr) {
		int ti=tind(ctr.getHWBase());
		s.append("ISR(TIMER");
		s.append(ti);
		s.append("_COMP");
		if(timer_max[ti]>255)
			s.append("A");
		s.append("_vect) {\n");
		s.append("\tadvanceCounter(");
		s.append(ctr.getName());
		s.append(");\n");
		s.append("}");
	}

	public void init(HighlevelConfiguration_Counter ctr) {
		long ns_interval = ctr.getTimeInNanoSeconds();
		int ti = tind(ctr.getHWBase());
		if(ns_interval < minNanoSeconds(null) || ns_interval > maxNanoSeconds(ti))
			throw new RuntimeException(ns_interval + "ns cannot be handled by hardware");
		counters_used[ti] = true;
		long reload, deviation;
		int pindex, pimax;

		reload = ns_interval * (long)arch.freq() / 1000000000L;
		for(pindex=0, pimax=timer_prescalers[ti].length-1; pindex <= pimax; pindex++) {
			long p = timer_prescalers[ti][pindex];
			if( (reload / p) < timer_max[ti] ) {
				if( (pindex==pimax) || ((reload % timer_prescalers[ti][pindex+1]) != 0) ) {
					break;
				}
			}
		}
		reload /= timer_prescalers[ti][pindex];
		assert reload<timer_max[ti];	

		cmdLine.verbln("Prescaler 1/" + timer_prescalers[ti][pindex]);
		cmdLine.verbln("OC " + reload);
		cmdLine.verbln("CPU freq " + arch.freq() + "Hz");


		deviation = reload * (long) timer_prescalers[ti][pindex] * 1000000000L / (long)arch.freq();
		deviation = ns_interval - deviation;
		cmdLine.verbln("Counter " +ctr.getHWBase() + " configured with deviation " +deviation +"ns");

		ocr[ti] = (int) reload;
		set_timsk(ctr,ti);
		set_tccr(ctr,ti,pindex);
	}

	protected void set_tccr(HighlevelConfiguration_Counter ctr, int ti, int prescalerIndex) {
		if(timer_max[ti] == 255) { // 8-bit timer
			tccr[ti] = 0x08 | (prescalerIndex+1);
		} else { // 16-bit timer
			tccr[ti] = 0x00;
			tccrb[ti] = 0x08 | (prescalerIndex+1);
		}
	}

	protected void set_timsk(HighlevelConfiguration_Counter ctr, int ti) {
		if(ti==2)
			timsk |= 0x80;
		else if(ti==1)
			timsk |= 0x10;
		else if(ti==0)
			timsk |=2;
		else if(ti==3)
			etimsk |= 0x10;
		else throw new
			RuntimeException("Interrupt Mask position unknown for timer " + ctr.getHWBase());
	}

	public void write_init(StringBuffer s) {
		for(int i=0; i<timer_max.length; i++) {
			if(! counters_used[i]) continue;
			if(timer_max[i]==255) { // 8-bit timer
				s.append("\tTCCR" + i + " = 0x" + Integer.toHexString(tccr[i]) + ";\n");
				s.append("\tOCR" + i + " = " + ocr[i] + ";\n");
				s.append("\tTCNT" + i + " = 0;\n");
			} else { // 16-bit timer
				s.append("\tTCCR" + i + "B = 0x" + Integer.toHexString(tccrb[i]) + ";\n");
				s.append("\tTCCR" + i + "A = 0x" + Integer.toHexString(tccr[i]) + ";\n");
				s.append("\tOCR" + i + "AH" + " = " + (ocr[i]>>>8)+";\n");
				s.append("\tOCR" + i + "AL" + " = " + (ocr[i]&0xff)+";\n");
				s.append("\tTCNT" + i + "H" + " = 0;\n");
				s.append("\tTCNT" + i + "L" + " = 0;\n");
			}
		}
		if(timsk != 0)
			s.append("\tTIMSK |= 0x" + Integer.toHexString(timsk) + ";\n");
		if(etimsk != 0)
			s.append("\tETIMSK |= 0x" + Integer.toHexString(etimsk) + ";\n");
	}

	private static ATMega __mcu;

	public static ATMega MCUDesc() {
		if(__mcu==null)
			throw new RuntimeException("MCU description not yet created");
		return __mcu;
	}

	public static ATMega MCUDesc(HighlevelConfiguration conf, CommandLineOptions clo, avr arch) {
		if(__mcu!=null) return __mcu;

		String mcuname = conf.getCodeHookVariable("mcu");
		if(mcuname==null)
			throw new RuntimeException("Define MCU type MCU to use hardware counter");
		try {
			__mcu = (ATMega) Class.forName("josek.arch.atmega." + mcuname).newInstance();
			__mcu.setCmdLineConf(clo,conf,arch);
			return __mcu;
		} catch (Exception e) {
			throw new RuntimeException("MCU " +mcuname+ " not implemented " + e.getMessage());
		}
	}

	/**
	 * Convert Timer name to index.
	 */
	protected static int tind(String timer_base) {
		if(timer_base.length() != 8 || !timer_base.toLowerCase().startsWith("counter"))
			throw new RuntimeException(timer_base + " is not a known counter");
		return Integer.parseInt(timer_base.substring(7,8));
	}

	private void setCmdLineConf(CommandLineOptions clo, HighlevelConfiguration co, avr arch) {
		conf=co;
		cmdLine=clo;
		this.arch = arch;
	}
}
