/**(c)

  Copyright (C) 2006, 2007, 2008, 2009 Michael Stilkerich, Johannes Bauer, Christian Wawersich 

  This file is part of the JOSEK operating system.

	JOSEK is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

	JOSEK is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with JOSEK. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package josek.idmap;

import josek.CommandLineOptions;
import josek.HighlevelConfiguration;

import java.util.Hashtable;
import java.util.Vector;
import java.util.TreeSet;
import java.util.Enumeration;
import java.util.Iterator;

public class IDMap<TC extends IDMapChildObject<TB>, TB extends IDMapBaseObject<TC>> {
	final CommandLineOptions cmdLine;
	final HighlevelConfiguration conf;

	private static final Vector<Mapper> mappers;

	private Hashtable<String, TB> baseObjects;
	private Hashtable<Integer, TB> baseObjectsByIID;
	private Hashtable<String, Vector<TC>> deferredMappings;
	private TreeSet<TB> baseObjectSet;
	private boolean finalized;
	private int minCPB;
	private int maxCPB;
	private int maxBaseID;
	private int maxChildID;
	private int childCount;
	private final int minBaseID;
	private String reverseMapMacro;
	private Mapper chosenMapper;

	private final String baseName;
	private final String childName;

	static {
		mappers = new Vector<Mapper>();
		mappers.add(new MapperIdentical());
		mappers.add(new MapperCombinedID());
	}

	public IDMap(int minBaseID, String baseName, String childName, CommandLineOptions clo, HighlevelConfiguration conf) {
		this.baseName = baseName;
		this.childName=childName;
		this.minBaseID = minBaseID;
		cmdLine=clo;
		this.conf=conf;
	
		baseObjects = new Hashtable<String, TB>();
		deferredMappings = new Hashtable<String, Vector<TC>>();
		baseObjectSet = new TreeSet<TB>();
		finalized = false;
		maxCPB=0;
		childCount=0;
		maxBaseID=0;
		maxChildID=0;
	}

	void setMaxID(int maxBaseID, int maxChildID) {
		this.maxBaseID = maxBaseID;
		this.maxChildID = maxChildID;
	}

	int getMinBaseID() { return minBaseID; }

	public int getMaxChildID() {
		chkFinalized();
		return maxChildID;
	}

	public int getMaxBaseID() {
		chkFinalized();
		return maxBaseID;
	}

	public String mapperName() {
		return baseName + " -> " + childName;
	}

	public String reverseMapMacroName() {
		return "Base"+baseName;
	}

	public boolean existsBase(String baseName) {
		return baseObjects.containsKey(baseName);
	}

	public void addBase(TB bobj) {
		if(baseObjects.containsKey(bobj.getName())) {
			throw new RuntimeException("Base Object " + bobj.getName() + " exists");
		}
		baseObjects.put(bobj.getName(), bobj);

		// add possibly deferred mappings
		if(deferredMappings.containsKey(bobj.getName())) {
			Vector<TC> children = deferredMappings.remove(bobj.getName());
			for(int i=0; i<children.size(); i++) {
				TC child = children.get(i);
				bobj.addChild(child);
				child.setBase(bobj);
			}
			if(bobj.numChildren() > maxCPB) {
				maxCPB = bobj.numChildren();
			}
		}
	}

	public void addChild(TC child, String baseName) {
		TB base = baseObjects.get(baseName);
		if(base == null) {
			Vector<TC> cobjs = deferredMappings.get(baseName);
			if(cobjs == null) {
				cobjs = new Vector<TC>();
				deferredMappings.put(baseName, cobjs);
			}
			cobjs.add(child);
		} else {
			base.addChild(child);
			child.setBase(base);
			if(base.numChildren() > maxCPB) {
				maxCPB = base.numChildren();
			}
		}
		childCount++;
	}

	private void chkFinalized() {
		if(!finalized) {
			throw new RuntimeException("ID Mapper " + mapperName() + " not finalized");
		}
	}

	public int getMaxCPB() { 
		chkFinalized();
		return maxCPB;
	}

	public int getMinCPB() {
		chkFinalized();
		return minCPB;
	}

	public Iterator<TB> baseIterator() {
		return baseObjectSet.iterator();
	}

	public int getBaseCount() {
		chkFinalized();
		return baseObjects.size();
	}

	public int getChildCount() {
		chkFinalized();
		return childCount; 
	}

	public void map() {
		if(finalized) {
			throw new RuntimeException("ID Mapper " + mapperName() + " already finalized");
		}

		if(deferredMappings.size() > 0) {
			throw new RuntimeException("ID Mapper " + mapperName() + " has pending deferred mappings");
		}
		deferredMappings=null;

		finalized = true;

		// Remove childless bases and generate sorted set
		minCPB = childCount;

		for(Enumeration<TB> bobje = baseObjects.elements();	bobje.hasMoreElements(); ) {
			TB bobj = bobje.nextElement();
			if(bobj.numChildren() == 0) {
				System.err.println("Removing childless " + baseName + " " + bobj.getName());
				baseObjects.remove(bobj.getName());
			} else {
				baseObjectSet.add(bobj);
				if(bobj.numChildren() < minCPB) {
					minCPB = bobj.numChildren();
				}
			}
		}

		// Choose the best mapper
		int cost=-1;
		for(int i=0; i<mappers.size(); i++) {
			Mapper m = mappers.get(i);
			int c = m.cost(this);
			if(c<0) continue;
			if(cost<0 || c<cost) {
				cost = c;
				chosenMapper=m;
				if(c==0) break;
			}
		}
		if(chosenMapper==null) {
			throw new RuntimeException("No suitable Mapper found for IDMap " + mapperName());
		}
		chosenMapper.map(this);

		// Create Hash to fetch base objects by internal id
		baseObjectsByIID = new Hashtable<Integer, TB>();
		for(Iterator<TB> it = baseObjectSet.iterator(); it.hasNext(); ) {
			TB bobj = it.next();
			baseObjectsByIID.put(new Integer(bobj.getInternalID()), bobj);

			// Debug out
			cmdLine.dbgln(bobj.toString());
		}

		// Set CodeHook Vars
		conf.setCodeHookVariable("max"+baseName, (new Integer(getMaxBaseID())).toString() );
		conf.setCodeHookVariable("max"+childName,(new Integer(getMaxChildID())).toString() );
	}

	public TB getBaseByIID(int iid) {
		chkFinalized();
		return baseObjectsByIID.get(new Integer(iid));
	}

	public void setReverseMapMacro(String macro) {
		if(reverseMapMacro != null) {
			throw new RuntimeException("Attempt to reset Reversemap Macro (was: " + reverseMapMacro + ")");
		}
		reverseMapMacro = macro;
	}

	public String getReverseMapMacro() {
		chkFinalized();
		return reverseMapMacro;
	}

	public TB getBase(String name) {
		return baseObjects.get(name);
	}

	public Enumeration<TC> childElements() {
		return new childEnumeration();
	}

	public String writeChildIndexedArray(String typeAndName, Getter<TC> gobj) {
		chkFinalized();
		return chosenMapper.writeChildIndexedArray(typeAndName, this, gobj);
	}

	private class childEnumeration implements Enumeration<TC> {
		private Iterator<TB> baseIter;
		private TB curBase;
		private int childIndex;

		private childEnumeration() {
			baseIter = baseIterator();
			childIndex=0;
			curBase=null;
		}

		public boolean hasMoreElements() {
			// all bases have at least one child, otherwise they'd have been killed during mapping
			if(baseIter.hasNext() || (curBase != null && curBase.numChildren() > childIndex))
				return true;
			return false;
		}

		public TC nextElement() throws java.util.NoSuchElementException {
			if(! hasMoreElements()) {
				throw new java.util.NoSuchElementException();
			}
			if(curBase != null && childIndex<curBase.numChildren()) {
				TC cld = curBase.getChild(childIndex);
				childIndex++;
				return cld;
			}
			childIndex=1;
			curBase = baseIter.next();
			return curBase.getChild(0);
		}
	}
}
