/**(c)

  Copyright (C) 2006, 2007, 2008, 2009 Michael Stilkerich, Johannes Bauer, Christian Wawersich 

  This file is part of the JOSEK operating system.

	JOSEK is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

	JOSEK is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with JOSEK. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package josek.idmap;

import java.util.Iterator;
import java.util.Enumeration;

/**
 * Creates a mapping that combines the ids of the base and the child elements
 * in a single value.
 *
 *  ------ Child ID -------
 * | BaseID     | Cld-Part |
 *  -----------------------
 * Costs of this mapping increases with an increasing divergence in the number
 * of child elements of the different base elements.
 */

final class MapperCombinedID implements Mapper {
	private int wastedBytes;

	MapperCombinedID() {
		wastedBytes = 0;
	}

	public int cost(IDMap idm) {
		int bitsneeded;
		int maxCPB = idm.getMaxCPB();

		// determine the # of bits needed for the child-ID portion
		bitsneeded = bitsNeeded(maxCPB);

		// feasible with 8-bit IDs?
		if( (1<<(8-bitsneeded)) < idm.getBaseCount() ) {
			return -1;
		}

		// evaluate RAM waste of Id Encoding
		wastedBytes = 0;
		for(Iterator<IDMapBaseObject> it = idm.baseIterator(); it.hasNext(); ) {
			IDMapBaseObject base = it.next();
			wastedBytes += (maxCPB-base.numChildren());
		}

		return wastedBytes;
	}

	private int bitsNeeded(int maxCPB) {
		int bitsneeded=0;

		if(maxCPB==0) return 0;

		for(int val=maxCPB-1; val>0; val>>>=1) {
			bitsneeded++;
		}
		return bitsneeded;
	}

	public void map(IDMap idm) {
		int baseInd=idm.getMinBaseID()-1;
		int baseid=0;
		int childid=0;
		int maxCPB = idm.getMaxCPB();
		int bitsneeded = bitsNeeded(maxCPB);

		for(Iterator<IDMapBaseObject> it = idm.baseIterator(); it.hasNext(); ) {
			baseInd++;
			baseid = baseInd << bitsneeded;
			IDMapBaseObject base = it.next();
			base.setInternalID(baseInd);

			for(int j=0; j<base.numChildren(); j++) {
				childid = baseid | j;
				base.getChild(j).setInternalID(childid);
			}
		}
		idm.setReverseMapMacro("#define "+idm.reverseMapMacroName()+"(cid) (cid>>"+bitsneeded+")");
		idm.cmdLine.verbln("CombinedID Encoding used for " + idm.mapperName() + " runtime mapping (" + wastedBytes + " wasted)");
		idm.setMaxID(baseid, childid);
	}

	public String writeChildIndexedArray(String typeAndName, IDMap idm, Getter gobj) {
		int i=0;
		StringBuffer s = new StringBuffer(typeAndName);
		s.append("[");
		s.append((idm.getMaxChildID()+1));
		s.append("] = { ");
		for(Enumeration<IDMapChildObject> e=idm.childElements(); e.hasMoreElements(); i++) {
			IDMapChildObject t = e.nextElement();
			while( i < t.getInternalID() ) {
				i++;
				s.append("0, ");
			}

			assert( i == t.getInternalID() );
			s.append(gobj.get(t));
			s.append(",");
		}
		s.replace(s.length()-1,s.length()," };");
		return s.toString();
	}
}

